# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
   rigid and affine transformation of image data and locations
"""
from __future__ import print_function, division
import sys
import math
import numpy as np
from scipy.spatial.transform import Rotation as R

from lcreg import image3d
from lcreg.lcreg_lib import interpolate_block
from lcreg.lcreg_util import system_memory_available_mb


def euler_from_matrix(m):
    """
        Calculate euler angles from matrix

        Args:
            m (numpy.array (4x4, float64) rotation matrix
        Returns:
            numpy.array: 3 euler angles
        """
    rot = R.from_matrix(m[:3, :3])
    return rot.as_euler("xyz", degrees=True)


def t_matrix_world_from_p(p):
    """
        Create a transformation matrix (world coordinates) from parameters p

        Args:
            p (numpy.array, float64): 6 (rigid) or 12 (affine) parameters
                rigid format: t_x, t_y, t_z, r_x, r_y, r_z
        Returns:
            numpy.array (4x4, float64): transformation matrix (voxel coords.)

    """
    m_world = np.identity(4, dtype=np.float64)
    if len(p) == 6:
        rot = R.from_rotvec(p[3:])
        m_world[:3, :3] = rot.as_matrix()
        m_world[:3, 3] = p[:3]
    else:
        m_world[:3, :4] = p.reshape((3, 4))
    return m_world


def t_matrix_vox_from_p(fixed_im, moving_im, p):

    """
        Create a transformation matrix (voxel coordinates) from parameters p

        Args:
            fixed_im (Image3D): fixed image
            moving_im (Image3D): moving image
            p (numpy.array, float64): 6 (rigid) or 12 (affine) parameters
        Returns:
            numpy.array (4x4, float64): transformation matrix (world coords.)

    """
    return t_matrix_world_to_vox(fixed_im, moving_im, t_matrix_world_from_p(p))


def p_from_t_matrix_world(M, p=None, is_affine=False):
    """
    Extract parameters from transformation matrix.

    Args:
        M (numpy.array, 4x3, numpy.float64): Transformation matrix
        p (numpy.array, 1x6 or 1x12, np.float64): parameter vector
            to fill in or None to create new vector
        is_affine (bool): indicates if transform ist affine or rigid
    Returns:
        numpy.array (1x6 or 1x12, numpy.float64): parameter vector

    """
    n_par = 12 if is_affine else 6
    if p is None:
        p = np.zeros(n_par, dtype=np.float64)
    else:
        if p.shape[0] != n_par:
            raise IndexError("wrong number of parameters for transform type")
    if is_affine:
        p[:] = M[:3, :].flatten()
    else:
        p[:3] = M[:3, 3]
        p[3:] = r_from_matrix(M)
    return p


def r_from_matrix(m):
    """
        calculate rotation vector from 3x3 matrix

        Args:
            m (numpy.array, float64): 3x3 rotation matrix
        Returns:
            np.array (1x3): rotation parameters (r_x, r_y, r_z)

    """

    rot = R.from_matrix(m[:3, :3])
    return rot.as_rotvec()


def swap_RAS_LPS(mat):
    """
    Switch beteween RAS and LPS system transformation

    Args:
        mat (np.array): transformation matrix in RAS/LPS system

    Returns:
        np.array: transformation matrix in LPS/RAS system

    """
    flip_xy = np.identity(4).astype(np.float64)
    flip_xy[0][0] = -1
    flip_xy[1][1] = -1
    return np.dot(flip_xy, np.dot(mat, flip_xy))


def t_matrix_world_to_vox(fixed_im, moving_im, m_world):
    """
    Transform transformation matrix from world to voxel coordinates.

    Args:
        fixed_im (image3d): fixed image (will be overwritten)
        moving_im (image3d): moving image (not modified)
        m_world (np.array, 4x4): transformation matrix fixed -> moving
                                 in world coordinates

    Returns:
        np.array 4x4: transformation matrix fixed -> moving in
                      voxel coordinates
    """
    return np.dot(
        moving_im.world_to_voxel_matrix,
        np.dot(m_world, fixed_im.voxel_to_world_matrix),
    ).astype(np.float64)


def transformed_offset_shape(fixed_im, moving_im, edges, m_world):
    """
    Calculate offset and shape of an image block

    Args:
        fixed_im (image3D): fixed image
        moving_im (image3D): moving image
        edges: (np.array, 3x8): edges of image block in voxel coordinates
        m_world(np.array, 4x4): transformation matrix in world coordinates

    Returns:
        tuple of tuples: offset and shape of the tranformed image region
                         order: xyz
    """
    m_vox = t_matrix_world_to_vox(fixed_im, moving_im, m_world)
    last_row = np.ones((1, edges.shape[1]))
    edges = np.vstack((edges, last_row))
    t_edges = np.dot(m_vox, edges)
    shape_array = moving_im.shape_array_xyz.astype(np.float64)
    t_offset = np.zeros(3, dtype=np.int32)
    t_shape = np.zeros(3, dtype=np.int32)
    for i in range(3):
        t_row = t_edges[i, :]
        t_min = int(min(t_row) + 0.5)
        if t_min <= 1:
            t_offset[i] = 0
        elif t_min > shape_array[i]:
            break
        else:
            t_offset[i] = t_min - 1
        t_max = int(max(t_row) + 0.5)
        if t_max > shape_array[i] - 2:
            t_shape[i] = shape_array[i] - t_offset[i]
        elif t_max < 0:
            break
        else:
            t_shape[i] = t_max - t_offset[i] + 2

    return (t_offset, t_shape)


def required_transform_mem_mb(s_in, fixed_im, moving_im, m_world):
    """
    Calculate required memory for image block transformation

    Args:
        s_in (tuple): size of image region in voxels, order xyz
        fixed_im (image3D): fixed image
        moving_im (image3D): moving image
        m_world(np.array, 4x4): transformation matrix in world coordinates

    Returns:
        int: Total memory required for fixed and moving image blocks in MB
    """
    n_voxels_in = s_in[0] * s_in[1] * s_in[2]
    nx, ny, nz = s_in
    edges_in = np.array(
        [
            [0, 0, 0],
            [nx, 0, 0],
            [0, ny, 0],
            [0, 0, nz],
            [0, ny, nz],
            [nx, 0, nz],
            [nx, ny, 0],
            [nx, ny, nz],
        ],
        dtype=np.float64,
    ).T
    # homogeneous coordinates
    last_row = np.ones((1, edges_in.shape[1]))
    edges_in = np.vstack((edges_in, last_row))
    m_vox = t_matrix_world_to_vox(fixed_im, moving_im, m_world)
    # transformed block edge coordinates
    edges_out = np.dot(m_vox, edges_in)[:3]
    out_min = np.min(edges_out, axis=1).astype(np.uint64) - 1
    out_max = np.max(edges_out, axis=1).astype(np.uint64) + 1
    n_voxels_out = int((out_max - out_min).prod() + 0.5)
    return (n_voxels_in + n_voxels_out) // 2 ** 18


def blockwise_grey_transform(
    in_img_name,
    out_img_name,
    m_world,
    mem_limit_mb=-1,
    n_threads=-1,
    is_mask=False,
):
    """
    Perform block wise geometric transformation of an image

    The transformation matrix m_world given in the world coordinate
    system is applied to in_img. The resampled image is stored in
    the bcolz directory given by out_img_name.

    Args:
        in_img_name (str): input bcolz base name
        out_img_name (str): output bcolz base name
        m_world (np.array): transformation matrix to be applied to in_img
        mem_limit_mb (int): memory limit for resampling in MB
        n_threads (int): number of threads to be used (-1: use all cores)
        is_mask (bool): True if image represents a binary mask

    """
    if mem_limit_mb < 0:
        mem_limit_mb = system_memory_available_mb()

    in_img = image3d.open_image(in_img_name, mode="r")
    out_img = image3d.empty_image_like(in_img, out_img_name)

    if 2 * in_img.mem_mb <= mem_limit_mb:
        in_data = in_img.data[:, :, :]
        out_data = out_img.data[:, :, :]
        t_matrix_vox = t_matrix_world_to_vox(out_img, in_img, m_world)
        interpolate_block(out_data, in_data, t_matrix_vox, n_threads, is_mask)
        out_img.data[:, :, :] = out_data[:, :, :]
        del in_data
        del out_data
    else:
        block_size = np.array((100, 100, 100), dtype=np.int32)
        mem_100 = required_transform_mem_mb(
            block_size, out_img, in_img, m_world
        )
        factor = math.pow(mem_limit_mb / mem_100, 1 / 3)
        block_size = (factor * block_size).astype(np.int32)
        for z in range(0, out_img.shape[0], block_size[2]):
            s_z = min(out_img.shape[0] - z, block_size[2])
            for y in range(0, out_img.shape[1], block_size[1]):
                s_y = min(out_img.shape[1] - y, block_size[1])
                for x in range(0, out_img.shape[2], block_size[0]):
                    s_x = min(out_img.shape[2] - x, block_size[0])
                    if min(s_x, s_y, s_z) > 0:
                        edges = np.array(
                            [
                                [x, y, z],
                                [x + s_x, y, z],
                                [x, y + s_y, z],
                                [x, y, z + s_z],
                                [x + s_x, y + s_y, z],
                                [x + s_x, y, z + s_z],
                                [x, y + s_y, z + s_z],
                                [x + s_x, y + s_y, z + s_z],
                            ],
                            dtype=np.float64,
                        ).T
                        in_offset, in_shape = transformed_offset_shape(
                            out_img, in_img, edges, m_world
                        )
                        if min(in_shape) > 0:
                            out_part = image3d.image_part(
                                out_img,
                                np.array((x, y, z)),
                                np.array((s_x, s_y, s_z)),
                            )
                            in_part = image3d.image_part(
                                in_img, in_offset, in_shape
                            )
                            t_matrix_vox = t_matrix_world_to_vox(
                                out_part, in_part, m_world
                            )
                            interpolate_block(
                                out_part.data,
                                in_part.data,
                                t_matrix_vox,
                                n_threads,
                                is_mask,
                            )
                            del in_part
                            out_img.data[
                                z : z + s_z, y : y + s_y, x : x + s_x
                            ] = out_part.data[:, :, :]
                            del out_part
    image3d.close_image(in_img)
    image3d.close_image(out_img)



#!/usr/bin/env python3

# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Convert Scanco ISQ file into mhd format without copying grey value data
"""

import os
import sys
from collections import OrderedDict
import numpy as np

# positions of relevant parameters in ISQ header. Numbers correspond
# to indices in a np.int32 array.
# Source: http://www.scanco.ch/en/support/customer-login/\
#                           faq-customers/faq-customers-general.html
__ISQ_OFFSETS_INT_4 = {
    "dimx_p": 11,
    "dimy_p": 12,
    "dimz_p": 13,
    "dimx_um": 14,
    "dimy_um": 15,
    "dimz_um": 16,
    "slice_thickness": 17,
    "slice_increment_um": 18,
    "slice_1_pos_um": 19,
    "min_data_value": 20,
    "max_data_value": 21,
    "mu_scaling": 22,
    "energy": 42,
    "intensity": 43,
    "data_offset": -1,
}

# Default values for mhd file parameters
__MHD_DEFAULTS = OrderedDict(
    [
        ("ObjectType", "Image"),
        ("NDims", "3"),
        ("BinaryData", "True"),
        ("BinaryDataByteOrderMSB", "False"),
        ("CompressedData", "False"),
        ("TransformMatrix", "1 0 0 0 1 0 0 0 1"),
        ("Offset", "0 0 0"),
        ("CenterOfRotation", "0 0 0"),
        ("AnatomicalOrientation", "RAI"),
        ("ElementSpacing", "1 1 1"),
        ("DimSize", "1 1 1"),
        ("HeaderSize", "-1"),
        ("ElementType", "MET_SHORT"),
    ]
)


def _read_isq_param(in_file_name):
    """
    Read parameters from ISQ file into OrderedDict

    Args:
        in_file_name (str): Input file name including suffix
    Returns:
        tuple of
            OrderedDict containing parameters in mhd style,
            offset of ISQ file data part in bytes
            grey value range (tuple)
    """
    param = __MHD_DEFAULTS.copy()
    isq_header = np.fromfile(in_file_name, np.int32, 128)
    # swap bytes if required
    if sys.byteorder == "big":
        isq_header = isq_header.byteswap()
    dim_p_str = ""
    element_spacing_str = ""
    # load image dimensions and calculate element spacing
    for i in range(3):
        dim_p = isq_header[__ISQ_OFFSETS_INT_4["dimx_p"] + i]
        dim_um = isq_header[__ISQ_OFFSETS_INT_4["dimx_um"] + i]
        dim_p_str += str(dim_p) + " "
        element_spacing_mm = dim_um / dim_p / 1000.0
        element_spacing_str += str(element_spacing_mm) + " "
    # get grey value range for windowing
    grey_min = isq_header[__ISQ_OFFSETS_INT_4["min_data_value"]]
    grey_max = isq_header[__ISQ_OFFSETS_INT_4["max_data_value"]]
    # set mhd parameters
    param["DimSize"] = dim_p_str[:-1]
    param["ElementSpacing"] = element_spacing_str[:-1]
    param["ElementType"] = "MET_SHORT"
    param["HeaderSize"] = "-1"
    param["ISQ_slice_thickness_um"] = str(
        isq_header[__ISQ_OFFSETS_INT_4["slice_thickness"]]
    )
    param["ISQ_slice_increment_um"] = str(
        isq_header[__ISQ_OFFSETS_INT_4["slice_increment_um"]]
    )
    param["ISQ_slice_1_pos_um"] = str(
        isq_header[__ISQ_OFFSETS_INT_4["slice_1_pos_um"]]
    )
    param["ISQ_min_data_value"] = str(grey_min)
    param["ISQ_max_data_value"] = str(grey_max)
    param["ISQ_mu_scaling"] = str(
        isq_header[__ISQ_OFFSETS_INT_4["mu_scaling"]]
    )
    param["ISQ_energy_V"] = str(isq_header[__ISQ_OFFSETS_INT_4["energy"]])
    param["ISQ_intensity_muA"] = str(
        isq_header[__ISQ_OFFSETS_INT_4["intensity"]]
    )
    param["ElementDataFile"] = os.path.basename(in_file_name)
    # calculate offset for grey value loading
    offset = (isq_header[__ISQ_OFFSETS_INT_4["data_offset"]] + 1) * 512
    grey_range = (grey_min, grey_max)
    return param, offset, grey_range


def isq_to_mhd(isq_file_name, mhd_file_name):
    """
        Convert ISQ file into meta image file
        ARGS:
            isq_file_name (str): full path name of isq image file
            mhd_file_name (str): name of mhd file to be written
    """
    mhd_param, offset, grey_range = _read_isq_param(isq_file_name)
    if os.sep in mhd_file_name:
        mhd_param["ElementDataFile"] = os.path.abspath(isq_file_name)
    else:
        mhd_param["ElementDataFile"] = isq_file_name
    with open(mhd_file_name, "w") as out_file:
        for i in mhd_param.items():
            out_file.write(i[0] + " = " + i[1] + "\n")


def main():
    if len(sys.argv) != 3:
        print("usage: isq_to_mhd isq_file_name mhd_file_name")
        sys.exit(0)
    else:
        isq_to_mhd(sys.argv[1], sys.argv[2])


if __name__ == "__main__":
    main()



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
main program for 3d image registration
"""
import sys
import logging
import configparser
import os
import ast
import multiprocessing
from sys import modules, argv
import time
import datetime
import shutil
import stat
import platform
import lcreg
import psutil

import numpy as np
import scipy
import bcolz

from lcreg import image3d
from lcreg import multires_pyramid
from lcreg.image_transform import (
    swap_RAS_LPS,
    p_from_t_matrix_world,
    t_matrix_world_from_p,
)
from lcreg import masking
from lcreg.optimisation import (
    simple_gradient,
    levenberg_marquardt,
    get_initial_transform_from_moments,
)
from lcreg.lcreg_util import system_memory_available_mb
from lcreg.lcreg_util import array_to_string
from lcreg.lcreg_util import image_corners_world
from lcreg.image_transform import euler_from_matrix


def create_output_directories(reg_config):
    """
    Create required output directories
    """
    for dir_name in (
        reg_config["DEFAULT"]["working_directory_name"],
        reg_config["DEFAULT"]["output_directory_name"],
    ):
        if not os.path.isdir(dir_name):
            try:
                os.makedirs(dir_name)
            except OSError:
                print(
                    'Could not create directory "' + dir_name + '" exiting ...'
                )
                exit(0)


def set_up_logging(reg_config, argv):
    """
    Initialise logging and print general information into log file
    """
    log_config = None
    log_level = reg_config["DEFAULT"]["log_level"]
    if log_level not in logging.__all__:
        raise ValueError("Illegal log level " + log_level)
    else:
        level_value = eval("logging." + log_level)
        if not isinstance(level_value, int):
            raise ValueError("Illegal log level " + log_level)
        else:
            log_format = "%(asctime)s - %(levelname)s: %(message)s"
            out_dir_name = reg_config["DEFAULT"]["output_directory_name"]
            log_filename = os.path.join(out_dir_name, "lcreg.log")
            log_config = {
                "format": log_format,
                "level": log_level,
                "filename": log_filename,
            }
            logging.basicConfig(**log_config)
            #
            # print system information
            logging.info("### lcreg " + argv[1])
            logging.info("")
            #
            # print CPU name if cpuinfo is installed
            try:
                import cpuinfo

                cpu_info = cpuinfo.get_cpu_info()
                logging.info("CPU: {}".format(cpu_info["brand_raw"]))
            except Exception:
                logging.info("CPU: unknown type")
            #
            # print maximum CPU frequency if value is available
            try:
                freq = psutil.cpu_freq()
                if freq is not None:
                    max_f = freq[2]
                    logging.info(
                        "CPU max. frequency: {} GHz".format(max_f / 1000)
                    )
            except Exception:
                logging.info("psutil: CPU max. frequency not accessible")
            #
            # nr of threads
            try:
                logging.info("CPU threads: {}".format(psutil.cpu_count()))
            except Exception:
                logging.info("psutil: CPU nr. of threads not accessible")
            #
            # print total physical memory if value is available
            try:
                mem = psutil.virtual_memory()
                if mem is not None:
                    mem_total_gb = mem[0] / 2 ** 30
                    logging.info("RAM: {:.1f} GiB".format(mem_total_gb))
            except Exception:
                logging.info("psutil: Amount of RAM not accessible")
            logging.info("")
            #
            # print Python and library versions
            logging.info(platform.platform())
            logging.info("Python {}".format(sys.version))
            logging.info("lcreg {}".format(lcreg.__version__))
            logging.info("numpy {}".format(np.__version__))
            logging.info("scipy {}".format(scipy.__version__))
            logging.info("bcolz {}".format(bcolz.__version__))
            logging.info("psutil {}".format(psutil.__version__))
            logging.info("")

    return log_config


def set_up_bcolz(reg_config):
    """
    Set global bcolz parameters
    """
    num_threads = reg_config.getint("DEFAULT", "nr_of_threads")
    if num_threads > 0:
        bcolz.blosc_set_nthreads(num_threads)
    bcolz.cparams.setdefaults(
        clevel=reg_config.getint("compression", "clevel"),
        shuffle=eval(reg_config["compression"]["shuffle"]),
        cname=reg_config["compression"]["cname"],
    )


def __import_image(t):
    """
    Import image

    Args:
        t (tuple): reg_config, name, mask_flag, log_config
    Returns:
        output name (str): name of bcolz image directory
    """
    reg_config, name, mask_flag, log_config = t
    # required for windows if multiple processes are used
    logging.basicConfig(**log_config)
    output_dir = reg_config["DEFAULT"]["working_directory_name"]
    output_name = os.path.join(output_dir, name)
    do_reuse = reg_config["DEFAULT"]["reuse_existing_input_images"] == "True"
    if do_reuse and os.path.isdir(output_name):
        logging.debug("reusing previously imported " + name + " image")
    else:
        grey_masking_range = None
        bits = None
        if not mask_flag:
            bits = reg_config.getint(
                "compression", name + "_image_bits_per_pixel"
            )
            range_str = reg_config["masking"][name + "_grey_masking_range"]
            if range_str != "None":
                values = range_str.split()
                low = ast.literal_eval(values[0])
                high = ast.literal_eval(values[1])
                grey_masking_range = (low, high)

        image3d.import_image(
            reg_config["DEFAULT"][name + "_image_name"],
            output_name,
            n_bits=bits,
            grey_range=grey_masking_range,
            is_mask=mask_flag,
        )
    return output_name


def __get_geometry(img_name):
    """
    Get geometry information of image

    Args:
        img_name (str): Name of bcolz image directory
    Returns:
        (tuple): shape, spacing, offset, transform_matrix
    """
    img = image3d.open_image(img_name)
    result = (
        img.shape_array_xyz,
        img.spacing,
        img.offset,
        img.transform_matrix,
    )
    image3d.close_image(img)
    return result


def __get_total_voxel_surface(img_name):
    """
    Get total voxel surface of image

    Args:
        img_name (str): Name of bcolz image directory
    Returns:
        (float): total voxel surface in mm^2
    """
    img = image3d.open_image(img_name)
    result = img.total_voxel_surface
    image3d.close_image(img)
    return result


def import_input_images(reg_config, log_config):
    """
    Import input images

    Args:
        rec_conig (ConfigParser): Configuration
    """
    logging.info("Image import started")
    arg_list = []
    arg_list.append((reg_config, "fixed", False, log_config))
    arg_list.append((reg_config, "moving", False, log_config))

    if reg_config["DEFAULT"]["fixed_mask_image_name"] != "None":
        arg_list.append((reg_config, "fixed_mask", True, log_config))
    if reg_config["DEFAULT"]["moving_mask_image_name"] != "None":
        arg_list.append((reg_config, "moving_mask", True, log_config))
    # import images in parallel only if not profiling
    if "profile" in modules or "cProfile" in modules:
        name_list = list(map(__import_image, arg_list))
    else:
        nr_of_threads = reg_config.getint("DEFAULT", "nr_of_threads")
        if nr_of_threads <= 0:
            nr_of_processes = min(multiprocessing.cpu_count(), len(arg_list))
        else:
            nr_of_processes = nr_of_threads
        pool = multiprocessing.Pool(nr_of_processes)
        name_list = pool.map(__import_image, arg_list)
        pool.close()
    #
    name_list.reverse()
    #
    # import fixed image
    fixed_name = name_list.pop()
    reg_config.set("import", "fixed_name", fixed_name)
    reg_config.set("pyramid", "fixed_grey_base", fixed_name)
    fixed_geometry = __get_geometry(fixed_name)
    if min(fixed_geometry[0]) < 6:
        logging.error("dimensions of fixed image too small")
        print("ERROR: dimensions of fixed image too small")
        exit()
    #
    # import moving image
    moving_name = name_list.pop()
    reg_config.set("import", "moving_name", moving_name)
    reg_config.set("pyramid", "moving_grey_base", moving_name)
    moving_geometry = __get_geometry(moving_name)
    if min(moving_geometry[0]) < 6:
        logging.error("dimensions of moving image too small")
        print("ERROR: dimensions of moving image too small")
        exit()
    #
    # import fixed mask and compare geometry with fixed_image
    if reg_config["DEFAULT"]["fixed_mask_image_name"] != "None":
        fixed_mask_name = name_list.pop()
        reg_config.set("import", "fixed_mask_name", fixed_mask_name)
        reg_config.set("pyramid", "fixed_mask_base", fixed_mask_name)
        fixed_mask_geometry = __get_geometry(fixed_mask_name)
        for i in range(len(fixed_mask_geometry)):
            if not np.allclose(fixed_geometry[i], fixed_mask_geometry[i]):
                logging.error("geometry of fixed mask and image differ")
                print("ERROR: geometry of fixed mask and image differ")
                exit()

    #
    # import moving mask and compare geometry with moving_image
    if reg_config["DEFAULT"]["moving_mask_image_name"] != "None":
        moving_mask_name = name_list.pop()
        reg_config.set("import", "moving_mask_name", moving_mask_name)
        reg_config.set("pyramid", "moving_mask_base", moving_mask_name)
        moving_mask_geometry = __get_geometry(moving_mask_name)
        for i in range(len(moving_mask_geometry)):
            if not np.allclose(moving_geometry[i], moving_mask_geometry[i]):
                logging.error("geometry of moving mask and image differ")
                print("ERROR: geometry of moving mask and image differ")
                exit()
    logging.info("Image import done")


def initialise_transform(reg_config):
    """
    Initialise tranformation

    Args:
        reg_config (ConfigParser): Confiuration
    """
    # check for illegal settings
    mat = np.identity(4, dtype=np.float64)
    in_file_name = reg_config["transformation"][
        "initial_transformation_file_name"
    ]
    if in_file_name != "None":
        if reg_config.getboolean(
            "transformation", "initialise_from_edge_moments"
        ):
            logging.error(
                "Two incompatible options for transfomation "
                + "initialisation chosen"
            )
            print(
                "ERROR: Two incompatible options for transfomation "
                + "initialisation chosen"
            )
            exit(0)
        if os.path.isfile(in_file_name):
            mat = np.loadtxt(in_file_name, dtype=np.float64)
        else:
            logging.error("could not find transformation file " + in_file_name)
    # convert RAS transform to LPS system
    if reg_config.getboolean("transformation", "initial_transform_is_RAS"):
        logging.info("converting initial transform from RAS to LPS system")
        mat = swap_RAS_LPS(mat)
    # invert initial transform if required
    if reg_config.getboolean(
        "transformation", "initial_transform_is_inverted"
    ):
        logging.info("inverting initial transform")
        mat = np.linalg.inv(mat)
    reg_config.set(
        "optimisation",
        "initial_transform",
        "np." + repr(mat).replace("=float64", "=np.float64"),
    )
    logging.debug("initial_transform: \n" + str(mat))


def create_multi_resolution_pyramid(reg_config):
    """
    Create multi resolution pyramid of fixed and moving images and masks

    Args:
        reg_config (ConfigParser): Configuration
    """
    multires_pyramid.create_pyramid(reg_config)


def perform_masking(reg_config):
    """
    Perform image masking for all resolutions available

    Args:
        reg_config (ConfigParser): Configuration
    """
    fixed_masked_images = []
    moving_masked_images = []
    nr_of_levels = len(
        ast.literal_eval(reg_config["pyramid"]["fixed_grey_files"])
    )
    fixed_gm_surface = reg_config.getfloat(
        "masking", "fixed_used_gradient_magnitude_fraction"
    ) * __get_total_voxel_surface(reg_config["import"]["fixed_name"])
    reg_config.set("masking", "fixed_gm_surface", str(fixed_gm_surface))
    moving_gm_surface = reg_config.getfloat(
        "masking", "moving_used_gradient_magnitude_fraction"
    ) * __get_total_voxel_surface(reg_config["import"]["moving_name"])

    reg_config.set("masking", "moving_gm_surface", str(moving_gm_surface))
    for l in range(nr_of_levels):
        out_files = masking.mask_level(reg_config, l)
        fixed_masked_images.append(out_files[0])
        moving_masked_images.append(out_files[1])

    reg_config["optimisation"]["fixed_masked_images"] = str(
        fixed_masked_images
    )
    reg_config["optimisation"]["moving_masked_images"] = str(
        moving_masked_images
    )


def initialise_from_moments(reg_config):
    """
    Initialise initial transformation from edge moments

    Args:
        reg_config (ConfigParser): Configuration
    """
    if reg_config.getboolean("transformation", "initialise_from_edge_moments"):
        get_initial_transform_from_moments(reg_config)


def perform_multi_resolution_registration(reg_config):
    """
    Multi resolution rigid or affine 3D registration

    Args:
        reg_config (ConfigParser): Configuration
    """
    logging.info("Multi resolution optimisation started")
    # get initial transformation parameters
    M_initial = eval(reg_config["optimisation"]["initial_transform"])
    is_affine = reg_config["transformation"]["transformation_type"] == "affine"
    p_initial = p_from_t_matrix_world(M_initial, None, is_affine)
    # get lists of image names
    fixed_images = ast.literal_eval(
        reg_config["optimisation"]["fixed_masked_images"]
    )
    moving_images = ast.literal_eval(
        reg_config["optimisation"]["moving_masked_images"]
    )

    p_opt = p_initial.copy()
    level = 0
    mem_limit_mb = ast.literal_eval(reg_config["DEFAULT"]["mem_limit_mb"])
    nr_of_threads = ast.literal_eval(reg_config["DEFAULT"]["nr_of_threads"])
    if mem_limit_mb < 0:
        mem_available = system_memory_available_mb()
    else:
        mem_available = min(system_memory_available_mb(), mem_limit_mb)

    # get levenberg_marquard parameters
    def get_config_LM(val_name):
        section = "levenberg_marquardt"
        return ast.literal_eval(reg_config[section][val_name])

    l_initial = float(get_config_LM("initial_lambda"))
    l_upscale_factor = get_config_LM("lambda_upscale_factor")
    l_downscale_factor = get_config_LM("lambda_downscale_factor")
    max_nr_of_iterations = get_config_LM("max_nr_of_iterations")
    max_nr_of_upscale_steps = get_config_LM("max_nr_of_upscale_steps")
    min_edge_voxel_displacement = get_config_LM("min_edge_voxel_displacement")
    p_opt = p_initial[:]

    level_results = []

    for fixed_name, moving_name in zip(
        fixed_images[::-1], moving_images[::-1]
    ):
        fixed_im = image3d.open_image(fixed_name)
        moving_im = image3d.open_image(moving_name)
        logging.info(
            "{:s} / {:s}".format(
                os.path.basename(fixed_name), os.path.basename(moving_name)
            )
        )
        if fixed_im.mem_mb + moving_im.mem_mb < 0.9 * mem_available:
            logging.info("  uncompressing images")
            fixed_im.uncompress()
            moving_im.uncompress()
        if reg_config.getboolean("simple_gradient", "active") and level == 0:
            delta_v_initial = ast.literal_eval(
                reg_config["simple_gradient"]["delta_v_initial"]
            )
            delta_v_min = ast.literal_eval(
                reg_config["simple_gradient"]["delta_v_min"]
            )
            c_rot = None
            v_rot = None
            if reg_config.getboolean(
                "transformation", "initialise_from_edge_moments"
            ):
                c_rot = eval(reg_config["optimisation"]["rotation_centre"])
                if not reg_config.getboolean(
                    "transformation", "initialise_shift_only"
                ):
                    v_rot = eval(reg_config["optimisation"]["rotation_axes"])

            result = simple_gradient(
                fixed_im,
                moving_im,
                p_opt,
                c_rot,
                v_rot,
                delta_v_initial,
                delta_v_min,
                mem_limit_mb,
                nr_of_threads,
            )
            logging.info(
                "simple grad., lcc_opt: {:f}, p_opt: {:s} n_steps: {:d}".format(
                    result[0], str(list(result[1])), result[2]
                )
            )
            p_opt = result[1].copy()

        result = levenberg_marquardt(
            fixed_im,
            moving_im,
            p_opt,
            l_initial,
            l_downscale_factor,
            l_upscale_factor,
            max_nr_of_iterations,
            min_edge_voxel_displacement,
            max_nr_of_upscale_steps,
            mem_limit_mb,
            nr_of_threads,
        )
        logging.info(
            "LM, lcc_opt: {:f}, p_opt: {:s}, n_steps: {:d}".format(
                result[0], str(list(result[1])), result[2]
            )
        )
        mat = t_matrix_world_from_p(result[1])
        logging.debug("m_opt:\n {:s}".format(str(mat)))
        logging.debug(
            "final euler angles (xyz, deg):\n" + str(euler_from_matrix(mat))
        )
        logging.debug("final translation vector:\n" + str(mat[:3, 3]))

        lcc_opt = result[0]
        p_opt = result[1][:]
        level_results.append((level, result))
        image3d.close_image(fixed_im)
        image3d.close_image(moving_im)
        level = level + 1
    reg_config.set(
        "optimisation",
        "level_results",
        repr(level_results)
        .replace("array", "np.array")
        .replace("float64", "np.float64"),
    )
    reg_config.set(
        "optimisation",
        "p_opt",
        repr(p_opt)
        .replace("array", "np.array")
        .replace("float64", "np.float64"),
    )
    reg_config.set("optimisation", "lcc_opt", str(lcc_opt))


def create_transformed_mhd(reg_config):
    """
    Create moving image mhd file with geometry information corresponding
        to fixed image

    Args:
        reg_config (ConfigParser): Configuration
    """

    # check if mhd transformation makes sense
    if reg_config["transformation"]["transformation_type"] != "rigid":
        return

    logging.info("create transformed mhd file")

    # get voxel to world matrix of moving image
    moving_name = reg_config["import"]["moving_name"]
    moving_image = image3d.open_image(moving_name, "r")
    moving_v2w_old = moving_image.voxel_to_world_matrix
    moving_spacing = moving_image.spacing

    # apply transform
    p_opt = eval(reg_config["optimisation"]["p_opt"])
    m_opt_inv = np.linalg.inv(t_matrix_world_from_p(p_opt))
    moving_v2w_new = np.dot(m_opt_inv, moving_v2w_old)
    # collect values
    new_offset = moving_v2w_new[0:3, 3]
    # correct for scaling
    new_transform_matrix = np.dot(
        moving_v2w_new[:3, :3], np.diag(1.0 / moving_spacing)
    )

    moving_mhd_name = reg_config["DEFAULT"]["moving_image_name"]
    moving_mhd = image3d.read_mhd_param(moving_mhd_name)

    # update mhd
    if os.path.isabs(moving_mhd["ElementDataFile"]):
        data_file_name = moving_mhd["ElementDataFile"]
    else:
        data_file_name = os.path.abspath(
            os.path.join(
                os.path.dirname(moving_mhd_name), moving_mhd["ElementDataFile"]
            )
        )

    moving_mhd["ElementDataFile"] = data_file_name
    moving_mhd["Offset"] = array_to_string(new_offset)
    # transpose rotation matrix for mhd export
    moving_mhd["TransformMatrix"] = array_to_string(new_transform_matrix.T)
    # store result
    mhd_out_name = os.path.join(
        reg_config["DEFAULT"]["output_directory_name"], "moving_registered.mhd"
    )
    image3d.write_mhd_file(moving_mhd, mhd_out_name)


def export_results(reg_config, script_suffix):
    """
    Store registration results

    Write transformation matrix, config file, c3d call script and
    transformation file according to RIRE. Perform moving image transformation
    if specified in reg_config

    Args:
        reg_config (ConfigParser): Configuration
        script_suffix: 'bat' or 'sh depending on os
    """
    # store matrix
    logging.info("storing transformation matrix")
    p_opt = eval(reg_config["optimisation"]["p_opt"])
    m_opt = t_matrix_world_from_p(p_opt)
    if reg_config.getboolean("transformation", "store_as_RAS"):
        m_opt_out = swap_RAS_LPS(m_opt)
    else:
        m_opt_out = m_opt
    out_file_name = os.path.join(
        reg_config["DEFAULT"]["output_directory_name"], "lcreg_result.mat"
    )
    np.savetxt(out_file_name, m_opt_out)
    #
    # store config file
    logging.info("storing config file")
    config_out_name = os.path.join(
        reg_config["DEFAULT"]["output_directory_name"], "lcreg_results.ini"
    )
    with open(config_out_name, "w") as out_file:
        reg_config.write(out_file)

    # store c3d call
    logging.info("storing c3d call file")
    c3d_call = "c3d "
    c3d_call += os.path.abspath(reg_config["DEFAULT"]["fixed_image_name"])
    c3d_call += " "
    c3d_call += os.path.abspath(reg_config["DEFAULT"]["moving_image_name"])
    c3d_call += " "
    c3d_call += "-reslice-matrix"
    c3d_call += " "
    c3d_call += os.path.abspath(
        os.path.join(
            reg_config["DEFAULT"]["output_directory_name"], "lcreg_result.mat"
        )
    )
    c3d_call += " "
    c3d_call += "-int Cubic -o"
    c3d_call += " "
    c3d_call += os.path.abspath(
        os.path.join(
            reg_config["DEFAULT"]["output_directory_name"],
            "resampled_moving_image_c3d.mhd",
        )
    )
    file_name = os.path.join(
        reg_config["DEFAULT"]["output_directory_name"],
        "c3d_call." + script_suffix,
    )
    with open(file_name, "w") as out_file:
        out_file.write(c3d_call)
    os.chmod(file_name, stat.S_IRWXU)

    if reg_config.getboolean("output", "resample_moving_image"):
        logging.info("resampling image with scipy.affine_transform")
        fixed_im = image3d.open_image(reg_config["import"]["fixed_name"], "r")
        moving_im = image3d.open_image(
            reg_config["import"]["moving_name"], "r"
        )
        # numpy array access order is z, y, x
        swap_xz = np.array(
            [[0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]],
            dtype=np.float64,
        )
        t_mat_vox = np.dot(fixed_im.voxel_to_world_matrix, swap_xz)
        t_mat_vox = np.dot(m_opt, t_mat_vox)
        t_mat_vox = np.dot(moving_im.world_to_voxel_matrix, t_mat_vox)
        t_mat_vox = np.dot(swap_xz.T, t_mat_vox)
        matrix = t_mat_vox[:3, :3]
        offset = t_mat_vox[:3, 3]
        image3d.close_image(fixed_im)
        image3d.close_image(moving_im)
        image3d.resample_mhd(
            reg_config["DEFAULT"]["fixed_image_name"],
            reg_config["DEFAULT"]["moving_image_name"],
            os.path.join(
                reg_config["DEFAULT"]["output_directory_name"],
                "resampled_moving_image_scipy.mhd",
            ),
            matrix,
            offset,
        )

    if reg_config.getboolean("output", "store_RIRE_validation_file"):
        logging.info("storing RIRE validation file")
        reg_config.set("output", "RIRE_date", time.strftime("%d %B %Y"))
        reg_config.set(
            "output",
            "RIRE_from",
            reg_config.get("DEFAULT", "fixed_modality").upper(),
        )
        reg_config.set(
            "output", "RIRE_to", reg_config.get("DEFAULT", "moving_modality")
        )
        reg_config.set(
            "output",
            "RIRE_patient_number",
            reg_config.get("DEFAULT", "patient_id")[-3:],
        )
        fixed_im = image3d.open_image(reg_config["import"]["fixed_name"], "r")
        R_mat = image_corners_world(fixed_im)
        image3d.close_image(fixed_im)
        R_mat_transformed = np.dot(m_opt, R_mat)
        combined_mat = np.hstack((R_mat[:3, :].T, R_mat_transformed[:3, :].T))
        out_file_name = (
            "RIRE_"
            + reg_config.get("DEFAULT", "patient_id")
            + "_"
            + reg_config.get("DEFAULT", "modality_combination")
            + ".transformation"
        )
        file_name = os.path.join(
            reg_config["DEFAULT"]["output_directory_name"], out_file_name
        )
        with open(file_name, "w") as out_file:
            out_file.write(reg_config.get("output", "RIRE_header"))
            out_file.write("\n\n")
            for row in range(8):
                row_str = "{:3d}  ".format(row + 1)
                for col in range(6):
                    row_str += "{:11.4f}".format(combined_mat[row, col])
                out_file.write(row_str + "\n")
            out_file.write("\n(All distances are in millimeters.)\n")
            out_file.write("-" * 73 + "\n")


def cleanup(reg_config):
    """
    Remove temporary files if specified in reg_config

    Args:
        reg_config (ConfigParser): Configuration
    """

    base_file_keys = (
        ("pyramid", "fixed_grey_base"),
        ("pyramid", "moving_grey_base"),
        ("pyramid", "fixed_mask_base"),
        ("pyramid", "moving_mask_base"),
    )

    pyramid_file_keys = (
        ("optimisation", "fixed_masked_images"),
        ("optimisation", "moving_masked_images"),
        ("pyramid", "fixed_grey_files"),
        ("pyramid", "moving_grey_files"),
        ("pyramid", "fixed_mask_files"),
        ("pyramid", "moving_mask_files"),
    )

    if reg_config.getboolean("DEFAULT", "remove_temporary_images"):
        for key in base_file_keys:
            if reg_config.has_option(key[0], key[1]):
                img_name = reg_config.get(key[0], key[1])
                if os.path.isdir(img_name):
                    shutil.rmtree(img_name)
        for key in pyramid_file_keys:
            if reg_config.has_option(key[0], key[1]):
                for img_name in ast.literal_eval(
                    reg_config.get(key[0], key[1])
                ):
                    if os.path.isdir(img_name):
                        shutil.rmtree(img_name)


def create_view_scripts(reg_config, script_suffix):
    """
    Create shell scripts for visualisation of registration results

    Args:
        reg_config (ConfigParser): Configuration
        script_suffix: 'bat' or 'sh' depending on os
    """
    fixed_image_name = reg_config.get("DEFAULT", "fixed_image_name")
    moving_image_name = reg_config.get("DEFAULT", "moving_image_name")
    output_directory_name = reg_config.get("DEFAULT", "output_directory_name")
    #
    # create snap commands for overlay
    if sys.platform.startswith("win"):
        snap_exe_name = "ITK-SNAP"
    else:
        snap_exe_name = "itksnap"
    # script filenames
    snap_unreg_call_file_name = os.path.join(
        output_directory_name, "snap_unreg." + script_suffix
    )
    snap_reg_call_file_name = os.path.join(
        output_directory_name, "snap_reg." + script_suffix
    )
    # write scripts
    with open(snap_unreg_call_file_name, "w") as out_file:
        out_file.write(
            snap_exe_name
            + " -g "
            + os.path.abspath(fixed_image_name)
            + " -o "
            + os.path.abspath(moving_image_name)
        )
    if reg_config["transformation"]["transformation_type"] == "rigid":
        with open(snap_reg_call_file_name, "w") as out_file:
            out_file.write(
                snap_exe_name
                + " -g "
                + os.path.abspath(fixed_image_name)
                + " -o "
                + os.path.abspath(
                    os.path.join(
                        output_directory_name, "moving_registered.mhd"
                    )
                )
            )
    #
    # create view_bcolz command for masked pyramids
    bcolz_call_file_name = os.path.join(
        output_directory_name, "view_masked_pyramid." + script_suffix
    )
    if not reg_config.getboolean("DEFAULT", "remove_temporary_images"):
        fixed_masked_images = ast.literal_eval(
            reg_config["optimisation"]["fixed_masked_images"]
        )
        moving_masked_images = ast.literal_eval(
            reg_config["optimisation"]["moving_masked_images"]
        )
        with open(bcolz_call_file_name, "w") as out_file:
            cmd = "view_compressed_images"
            cmd = cmd + " " + os.path.abspath(output_directory_name)
            for d_name in fixed_masked_images + moving_masked_images:
                cmd = cmd + " " + os.path.abspath(d_name)
            out_file.write(cmd)
    # set execute flag on linux
    if not sys.platform.startswith("win"):
        for file_name in (
            bcolz_call_file_name,
            snap_unreg_call_file_name,
            snap_reg_call_file_name,
        ):
            if os.path.isfile(file_name):
                os.chmod(file_name, stat.S_IRWXU)


def main(parameter_file_name=None):
    """
    main function

    Args:
        parameter_file_name (str): Parameter file name
    """
    start_time = time.time()
    #
    # check command line argument
    if parameter_file_name is None and len(argv) != 2:
        print("\nusage: lcreg parameter_file_name\n")
        exit(0)
    else:
        parameter_file_name = argv[1]
    #
    # read config file
    reg_config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation()
    )
    try:
        reg_config.read_file(open(parameter_file_name))
    except FileNotFoundError:
        print(
            'Configuration file "'
            + parameter_file_name
            + '" not found, exiting.'
        )
        exit(0)
    #
    # add new sections
    reg_config.add_section("import")
    reg_config.add_section("pyramid")
    reg_config.add_section("optimisation")
    #
    create_output_directories(reg_config)
    #
    log_config = set_up_logging(reg_config, argv)
    #
    set_up_bcolz(reg_config)
    #
    import_input_images(reg_config, log_config)
    #
    initialise_transform(reg_config)
    #
    create_multi_resolution_pyramid(reg_config)
    #
    perform_masking(reg_config)
    #
    initialise_from_moments(reg_config)
    #
    perform_multi_resolution_registration(reg_config)
    #
    create_transformed_mhd(reg_config)
    #
    # set suffix for sh/bat files
    script_suffix = "bat" if sys.platform.startswith("win") else "sh"
    #
    export_results(reg_config, script_suffix)
    #
    create_view_scripts(reg_config, script_suffix)
    #
    cleanup(reg_config)
    #
    end_time = time.time()
    delta_t = datetime.timedelta(seconds=end_time - start_time)
    logging.info("*** lcreg stopped after {}\n".format(delta_t))


def profiling_main():
    import pstats
    import cProfile

    #
    # check command line argument
    if len(sys.argv) != 2:
        print("\nusage: python lcreg_profile parameter_file_name\n")
        sys.exit(0)
    #
    # call registration
    cProfile.runctx("main()", globals(), locals(), "lcreg.prof")
    #
    stats = pstats.Stats("lcreg.prof")
    stats.strip_dirs().sort_stats("cumtime").print_stats()


if __name__ == "__main__":
    # call registration
    main()



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
    Multiresolution representation of 3d images.
"""

from __future__ import print_function, division
import logging
import math
import os.path
import numpy as np

from lcreg import image3d
from lcreg import lcreg_lib as cy_ip

from lcreg.lcreg_util import system_memory_available_mb

_SQRT2 = math.sqrt(2)
_SCALE_SQRT2 = np.array([_SQRT2, _SQRT2, _SQRT2], dtype=np.float64)
_SCALE_2 = np.array([2, 2, 2], np.float64)


def _resampling_mem_needed_mb(out_shape, scale_factor):
    n_out = np.array(out_shape, dtype=np.uint64)[::-1]
    n_in = np.ceil((n_out + 1) * scale_factor).astype(np.uint64)
    mem_needed = (
        2 * (n_in.prod() + n_out.prod()) + 4 * n_out[0] * n_out[1] * n_in[2]
    )
    return mem_needed // 2 ** 20


def _blockwise_resample(
    in_im_name,
    out_im_name,
    scale_factor,
    is_mask=False,
    mem_limit_mb=-1,
    thread_num=-1,
):
    """
        Resample an image to another resolution.

        This function resamples an image block by block to another
        resolution using not more than the specified amount of RAM.

        Args:
            in_im_name (str): name of input image
            out_im_name (str): name of output image
            scale_factor (numpy array): 3d scaling factor for
                voxel size. Thus, a scaling factor > 1 corresponds
                to downsampling. Note: order is (x, y, z)
            is_mask (bool): flag switching between grey value and
                mask mode
            mem_limit_mb (int): upper memory limit for internal
                buffers in mb. A value of -1 indicates that all
                available RAM should be used.
            thread_num (int): number of threads to be used for
                the calculation. A value of -1 indicates that all
                available cpu cores should be used
    """
    # open input image
    in_im = image3d.open_image(in_im_name, mode="r")
    # calculate output image size
    out_im_shape_array_xyz = np.round(
        in_im.shape_array_xyz / scale_factor
    ).astype(np.uint32)
    out_im_shape = tuple(out_im_shape_array_xyz[::-1])
    # correct scaling factor w.r.t rounding
    scale_factor = in_im.shape_array_xyz / out_im_shape_array_xyz
    scale_factor = scale_factor.astype(np.float64)
    # spacing of output image
    out_im_spacing = in_im.spacing * scale_factor
    # logging output
    logging.info("  spacing / mm: " + str(list(out_im_spacing)))
    logging.info("  size: " + str(out_im_shape))
    # calculate orign of output image
    out_im_offset_vox = 0.5 * scale_factor - 0.5
    out_im_offset = in_im.transform_voxel_to_world(out_im_offset_vox)

    out_im = image3d.create_image(
        out_im_name,
        out_im_shape,
        out_im_offset,
        out_im_spacing,
        in_im.transform_matrix,
        mode="w",
    )
    out_im.data.flush()

    # get amount of memory to be used
    if mem_limit_mb < 0:
        mem_limit_mb = system_memory_available_mb()

    # calculated memory required for downsampling
    mem_needed_mb = _resampling_mem_needed_mb(out_im.shape, scale_factor)

    # perform resampling
    if mem_needed_mb < mem_limit_mb:
        in_array = in_im.data[:, :, :]
        out_array = out_im.data[:, :, :]
        cy_ip.downsample(in_array, out_array, thread_num, is_mask)
        out_im.data[:, :, :] = out_array[:, :, :]
    else:
        batch_size_z = 1
        while (
            _resampling_mem_needed_mb(
                (batch_size_z, out_im.shape[1], out_im.shape[2]), scale_factor
            )
            < mem_limit_mb
        ):
            batch_size_z += 1
        batch_size_z -= 1
        if batch_size_z == 0:
            raise MemoryError("Insufficient Memory for downsampling")
        # allocate temporary array
        tmp_shape_z = int(np.ceil((batch_size_z + 2) * scale_factor[2]))
        tmp_shape = (tmp_shape_z, out_im.shape[1], out_im.shape[2])
        tmp_array = np.zeros(shape=tmp_shape, dtype=np.float64)
        for z_pos in range(0, out_im.shape[0], batch_size_z):
            tmp_array.fill(0.0)
            z_end = min(z_pos + batch_size_z, out_im.shape[0])
            out_block = out_im.data[z_pos:z_end, :, :]
            in_start = max(int(np.floor(z_pos * scale_factor[2] - 1)), 0)
            in_end = min(
                int(np.ceil((z_end + 1) * scale_factor[2] - 1)), in_im.shape[0]
            )
            in_block = in_im.data[in_start:in_end, :, :]
            in_offset_z = z_pos * scale_factor[2] - in_start
            in_offset = np.array([0, 0, in_offset_z], dtype=np.float64)
            cy_ip.downsample_with_offset(
                in_block,
                out_block,
                tmp_array[0 : in_end - in_start, :, :],
                scale_factor,
                in_offset,
                thread_num,
                is_mask,
            )
            out_im.data[z_pos:z_end, :, :] = out_block[:, :, :]
        del in_block
        del out_block
        del tmp_array

    # close images
    image3d.close_image(in_im)
    image3d.close_image(out_im)


def _create_image_pyramid(
    base_img_name,
    highest_resolution,
    is_sqrt2_pyramid,
    nr_of_levels,
    is_mask,
    mem_limit_mb,
    thread_num,
):
    """
        Args:
            base_img_name: name of image the pyramid is based on
            nr_of_levels (int): number of levels to be generated
            highest_resolution (np.array): highest voxel spacing
                in the pyramid (mm)
            is_mask (bool): flag switching between grey value and
                mask mode
            mem_limit_mb (int): upper memory limit for internal
                buffers in mb. A value of -1 indicates that all
                available RAM should be used.
            thread_num (int): number of threads to be used for
                the calculation. A value of -1 indicates that all
                available cpu cores should be used
        Returns:
            list: image file names starting with highest resolutions
    """
    file_name_list = []
    base_img = image3d.open_image(base_img_name)
    base_spacing = base_img.spacing
    image3d.close_image(base_img)

    # resample base image if required
    if np.allclose(base_img.spacing, highest_resolution):
        file_name_list.append(base_img_name)
    else:
        scale_factor = highest_resolution / base_spacing
        out_img_name = base_img_name + "_l_00"
        logging.info(os.path.basename(out_img_name) + " :")
        _blockwise_resample(
            base_img_name,
            out_img_name,
            scale_factor,
            is_mask,
            mem_limit_mb,
            thread_num,
        )
        file_name_list.append(out_img_name)

    # perform downsampling
    for i in range(1, nr_of_levels + 1):
        if is_sqrt2_pyramid:
            if i == 1:
                scale_factor = _SCALE_SQRT2
                in_img_name = file_name_list[0]
            else:
                scale_factor = _SCALE_2
                in_img_name = file_name_list[i - 2]
        else:
            scale_factor = _SCALE_2
            in_img_name = file_name_list[i - 1]

        out_img_name = "%s_l_%02i" % (base_img_name, i)
        logging.info(os.path.basename(out_img_name) + " :")
        _blockwise_resample(
            in_img_name,
            out_img_name,
            scale_factor,
            is_mask,
            mem_limit_mb,
            thread_num,
        )
        file_name_list.append(out_img_name)
    return file_name_list


def create_pyramid(reg_config):
    """
       Create multi resolution pyramid according to configuration file.

       Args:
        reg_config (configparser.ConfigParser): regpy3d configuration
    """
    logging.info("Started pyramid generation")
    base_spacing = None
    # determine finest resolution
    for n in [
        "fixed_grey_base",
        "moving_grey_base",
        "fixed_mask_base",
        "moving_mask_base",
    ]:
        if reg_config.has_option("pyramid", n):
            img = image3d.open_image(reg_config["pyramid"][n])
            if base_spacing is None:
                base_spacing = img.spacing
            else:
                base_spacing = np.max([base_spacing, img.spacing], axis=0)
            image3d.close_image(img)
    if reg_config.getboolean("multiresolution", "pyramid_cubic_voxels"):
        base_spacing.fill(np.max(base_spacing))
    base_spacing *= reg_config.getfloat(
        "multiresolution", "pyramid_min_scale_factor"
    )
    logging.debug("base_spacing: {} mm".format(base_spacing))
    pyramid_sqrt2_scaling = reg_config.getboolean(
        "multiresolution", "pyramid_sqrt2_scaling"
    )
    # get scale factor
    if pyramid_sqrt2_scaling:
        level_scale_factor = _SQRT2
    else:
        level_scale_factor = float(2)
    # determine number of levels
    level_nr = None
    min_vox_nr = eval(reg_config["multiresolution"]["pyramid_min_voxel_nr"])
    for n in [
        "fixed_grey_base",
        "moving_grey_base",
        "fixed_mask_base",
        "moving_mask_base",
    ]:
        if reg_config.has_option("pyramid", n):
            img = image3d.open_image(reg_config["pyramid"][n])
            factor = np.prod(img.spacing) / np.prod(base_spacing)
            nr_vox_at_base_level = img.nr_of_voxels * factor
            current_level_nr = int(
                math.log(nr_vox_at_base_level / min_vox_nr)
                / math.log(level_scale_factor)
                / 3.0
                + 0.5
            )
            # need at least 7 as mininum nr of voxels in each direction
            dim_level_nr = int(
                math.log(min(img.shape) / 7) / math.log(level_scale_factor)
            )
            if level_nr is None:
                level_nr = min(current_level_nr, dim_level_nr)
            else:
                level_nr = min(level_nr, current_level_nr, dim_level_nr)
            image3d.close_image(img)
    logging.debug("nr of levels: {}".format(level_nr))
    # create levels
    mem_limit_mb = reg_config.getint("DEFAULT", "mem_limit_mb")
    nr_of_threads = reg_config.getint("DEFAULT", "nr_of_threads")
    for n in [
        "fixed_grey_base",
        "moving_grey_base",
        "fixed_mask_base",
        "moving_mask_base",
    ]:
        if reg_config.has_option("pyramid", n):
            im_name = reg_config["pyramid"][n]
            file_name_list = _create_image_pyramid(
                im_name,
                base_spacing,
                pyramid_sqrt2_scaling,
                level_nr,
                "mask" in n,
                mem_limit_mb,
                nr_of_threads,
            )
            list_name = n.replace("base", "files")
            reg_config.set("pyramid", list_name, str(file_name_list))
    logging.info("Pyramid generation done")



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
    Creation of masked images. Masking is based on gradient magnitude and
    optionally on binary images provided by the user,
"""

import logging
import ast
import multiprocessing
import os.path
import numpy as np
import bcolz

from lcreg import image3d
from lcreg import lcreg_lib as cy_ip
from lcreg import lcreg_util


def _fill_gm_buf(in_data, mask_data, gm_buf, block_size, thread_num=-1):
    gm_max = -1.0
    for z in range(0, in_data.shape[0] - 1, block_size):
        if z == 0:
            # keep first gm slice empty
            gm_offset_low = 1
        else:
            # use first gm slice
            gm_offset_low = 0
        if z + block_size >= in_data.shape[0]:
            # calculate number of remaining gm rows
            # last row is left empty
            gm_offset_high = in_data.shape[0] - z - 2
        else:
            # use block_size rows
            gm_offset_high = block_size

        in_block = in_data[
            z + gm_offset_low - 1 : z + gm_offset_high + 1, :, :
        ]
        gm_block = gm_buf[z + gm_offset_low : z + gm_offset_high, :, :]
        if mask_data is None:
            mask_block = None
        else:
            mask_block = mask_data[
                z + gm_offset_low : z + gm_offset_high, :, :
            ]
        cy_ip.fill_gm_image(in_block, mask_block, gm_block, thread_num)
        gm_buf[z + gm_offset_low : z + gm_offset_high, :, :] = gm_block[
            :, :, :
        ]
        gm_max = max(gm_max, cy_ip.gm_max(gm_block))
    gm_buf.flush()
    return gm_max


def _create_gm_histogram(gm_buf, gm_max, block_size):
    gm_hist = np.zeros(shape=(2 ** 20,), dtype=np.uint64)
    for z in range(1, gm_buf.shape[0] - 1, block_size):
        gm_block = gm_buf[z : min(z + block_size, gm_buf.shape[0] - 1), :, :]
        cy_ip.fill_gm_histogram(gm_block, gm_hist, gm_max)
    return gm_hist


def _fill_masked_image(
    in_data, out_data, gm_buf, gm_limit, is_moving, block_size, thread_num=-1
):
    for z in range(1, gm_buf.shape[0] - 1, block_size):
        gm_block = gm_buf[z : min(z + block_size, gm_buf.shape[0] - 1), :, :]
        z_low_img = max(0, z - 2)
        z_high_img = min(z + block_size + 2, in_data.shape[0])
        in_block = in_data[z_low_img:z_high_img, :, :]
        out_block = out_data[z_low_img:z_high_img, :, :]
        out_data[z_low_img:z_high_img, :, :] = out_block[:, :, :]
        z_offset = z - z_low_img
        cy_ip.gm_masking(
            in_block, out_block, gm_block, z_offset, gm_limit, is_moving
        )
        out_data[z_low_img:z_high_img, :, :] = out_block[:, :, :]
    if not is_moving:
        for z in range(1, out_data.shape[0] - 1, block_size):
            z_max = min(z + block_size, out_data.shape[0] - 1)
            out_block = out_data[z:z_max, :, :]
            cy_ip.do_run_length_encoding(out_block, thread_num)
    out_data.flush()


def _mask_image(
    in_im_name, mask_name, gm_surface, mem_limit_mb, thread_num, out_im_name
):
    """
        Create mask image in internal format

        Args:
            in_im_name (str): Name of input image rootdir
            in_mask_name (str): Name of input mask rootdir
            gm_surface (float): desired gm surface in mm^2
            mem_limit_mb (int): Amount of memory to use for masking
            thread_num (int): Nr. of threads
            out_im_name (int): Name of output grey image
            remove_temporary_images (bool): remove temporary images if True
    """
    # initialise thread_num and mem limit
    if thread_num <= 0:
        thread_num = lcreg_util.system_nr_of_cores()
    if mem_limit_mb <= 0:
        mem_limit_mb = lcreg_util.system_memory_available_mb()
    # open/create images and initialise data reference
    # input image
    in_im = image3d.open_image(in_im_name, mode="r")
    in_data = in_im.data
    # calculate gm_fraction
    gm_fraction = gm_surface / in_im.total_voxel_surface
    # one thread per image line
    thread_num = min(thread_num, in_im.shape[1])
    # output image
    out_im = image3d.empty_image_like(in_im, out_im_name)
    out_data = out_im.data
    # mask image
    if mask_name != "None":
        mask_im = image3d.open_image(mask_name, mode="r")
        mask_data = mask_im.data
    else:
        mask_im = None
        mask_data = None
    #
    # determine block size
    bytes_per_voxel = 2 + 4 + 4
    if mask_im is not None:
        bytes_per_voxel += 2
    bytes_per_slice = (
        bytes_per_voxel * int(in_im.shape[1]) * int(in_im.shape[2])
    )
    block_size = min(in_im.shape[0], mem_limit_mb * 2 ** 20 // bytes_per_slice)
    #
    # allocate float buffer for gm values
    buf_slice = np.zeros(
        shape=(1, in_im.shape[1], in_im.shape[2]), dtype=np.float32
    )
    gm_buf_size_mb = in_im.mem_mb * 2
    if gm_buf_size_mb < image3d.BCOLZ_COMPRESSION_LIMIT_MB:
        c_params = bcolz.cparams(clevel=1, shuffle=bcolz.NOSHUFFLE)
    else:
        c_params = None
    gm_buf_name = in_im_name + "_gm_buf"
    gm_buf = bcolz.carray(
        buf_slice,
        rootdir=gm_buf_name,
        cparams=c_params,
        expectedlen=in_im.nr_of_voxels,
        mode="w",
    )
    for _ in range(in_im.shape[0] - 1):
        gm_buf.append(buf_slice)
    gm_buf.flush()
    #
    # fill float image and calculate maximum gm value
    gm_max = _fill_gm_buf(in_data, mask_data, gm_buf, block_size, thread_num)
    #
    # reopen gm_buf in read only mode
    gm_buf.flush()
    del gm_buf
    gm_buf = bcolz.open(gm_buf_name, mode="r")
    #
    # close mask image
    if mask_im is not None:
        image3d.close_image(mask_im)
    #
    # create gm_histogram
    hist = _create_gm_histogram(gm_buf, gm_max, block_size)
    #
    # calculate gm_limit
    n_elem = np.sum(hist, dtype=np.uint64)
    count_limit = int(n_elem * (1 - gm_fraction) + 0.5)
    cum_hist = np.cumsum(hist)
    bin_index = np.argmin(np.abs(cum_hist - count_limit))
    gm_limit = gm_max * bin_index / hist.shape[0]
    #
    #
    # fill in the output image
    is_moving = "moving" in out_im_name
    _fill_masked_image(
        in_data, out_data, gm_buf, gm_limit, is_moving, block_size, thread_num
    )
    logging.debug(
        os.path.basename(out_im_name)
        + " compression ratio: "
        + str(out_data.nbytes / out_data.cbytes)
    )
    # cleanup
    # delete gm buffer
    gm_buf.purge()
    # close images
    image3d.close_image(in_im)
    image3d.close_image(out_im)


def mask_level(reg_config, level):
    """
        Create masked images for a certain level.

        Args:
            reg_config (ConfigParser): Registration configuration
            level (int): level for which masked image is to be created

        Returns:
            tuple(str, str): Names of created fixed and moving masked image
    """
    mem_limit_mb = reg_config.getint("DEFAULT", "mem_limit_mb")
    # memory available
    if mem_limit_mb < 0:
        mem_limit_mb = lcreg_util.system_memory_available_mb()
    # number of threads to use
    thread_num = reg_config.getint("DEFAULT", "nr_of_threads")
    if thread_num < 0:
        thread_num = multiprocessing.cpu_count()
    #
    # mask fixed image
    in_im_name = ast.literal_eval(reg_config["pyramid"]["fixed_grey_files"])[
        level
    ]
    if reg_config["DEFAULT"]["fixed_mask_image_name"] != "None":
        mask_name = ast.literal_eval(
            reg_config["pyramid"]["fixed_mask_files"]
        )[level]
    else:
        mask_name = "None"
    gm_surface = reg_config.getfloat("masking", "fixed_gm_surface")
    fixed_out_im_name = in_im_name + "_masked"
    _mask_image(
        in_im_name,
        mask_name,
        gm_surface,
        mem_limit_mb,
        thread_num,
        fixed_out_im_name,
    )
    #
    # mask moving image
    in_im_name = ast.literal_eval(reg_config["pyramid"]["moving_grey_files"])[
        level
    ]
    if reg_config["DEFAULT"]["moving_mask_image_name"] != "None":
        mask_name = ast.literal_eval(
            reg_config["pyramid"]["moving_mask_files"]
        )[level]
    else:
        mask_name = "None"
    gm_surface = reg_config.getfloat("masking", "moving_gm_surface")
    moving_out_im_name = in_im_name + "_masked"
    _mask_image(
        in_im_name,
        mask_name,
        gm_surface,
        mem_limit_mb,
        thread_num,
        moving_out_im_name,
    )
    #
    # return names of created images
    return (fixed_out_im_name, moving_out_im_name)



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

#!/usr/bin/env python

"""
convert compressed image to mhd and view the result with itkSNAP
"""
import sys
import os
from subprocess import call
import tempfile
import shutil
from lcreg import image3d


def main():
    #
    # usage message
    if len(sys.argv) < 3:
        print("usage: view_compressed_images tmp_prefix path_name_1 ", end="")
        print(" path_name_2 ...")
        return
    #
    # check for valid temporary path
    root_tmp_path = sys.argv[1]
    if root_tmp_path[-1] != os.path.sep:
        root_tmp_path += os.path.sep
    tmp_path = tempfile.mkdtemp(prefix=root_tmp_path)
    #
    # export bcolz images to mhd format and remember mhd file names
    mhd_file_names = []
    for bcolz_name in sys.argv[2:]:
        while bcolz_name[-1] == os.path.sep:
            bcolz_name = bcolz_name[:-1]
        mhd_file_names.append(
            os.path.join(tmp_path, os.path.basename(bcolz_name) + ".mhd")
        )
        try:
            image3d.export_bcolz_image(bcolz_name, mhd_file_names[-1], True)
        except Exception as e:
            print(str(e))
            return
    #
    # set up command and call it
    if sys.platform == "darwin":
        MAC_OS_SNAP_ENVIRON = "ITKSNAP_APP_DIR"
        if MAC_OS_SNAP_ENVIRON in os.environ:
            snap_exe = (
                os.environ[MAC_OS_SNAP_ENVIRON]
                + os.sep
                + "ITK-SNAP.app/Contents/MacOS/ITK-SNAP"
            )
        else:
            print(
                f"\nRequired environment variable {MAC_OS_SNAP_ENVIRON}"
                " not set."
            )
            print("Exiting ...\n")
            sys.exit(-1)
    elif sys.platform.startswith("win"):
        snap_exe = "ITK-SNAP"
    else:
        snap_exe = "itksnap"

    cmd = snap_exe + " -g " + mhd_file_names[0]
    if len(mhd_file_names) > 1:
        cmd += " -o"
    for fn in mhd_file_names[1:]:
        cmd += " " + fn
    call(cmd.split())
    #
    # cleanup
    shutil.rmtree(tmp_path)


if __name__ == "__main__":
    main()



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Module defining the Image3D class, Image conversion and I/O functionality.
Internally, images are represented in terms of bcolz arrays using compression.
The internal representation can be converted into meta image files (mhd).
"""
from __future__ import print_function, division
import sys
import os
import gc
import time
import logging
from collections import OrderedDict
import numpy as np
import bcolz
from lcreg.lcreg_lib import remove_rle, rescale_gvals
from lcreg.lcreg_util import array_to_string, string_to_array

BCOLZ_COMPRESSION_LIMIT_MB = 128


class Image3D(object):
    """
        Image3D class containing image data (numpy or on-disk bcolz array)
        and geometry. Essential properties required for the
        creation of meta image files (mhd) are stored in bcolz.attrs.
    """

    def __init__(
        self, in_data, offset=None, spacing=None, transform_matrix=None
    ):
        """
        Image3D constructor.

        Initialises Image3D geometry using the passed offset, spacing
        and transform matrix (rotation).

        Args:
           in_data (bcolz.carray or numpy array): grey value data.
           offset (numpy array): 3D image offset in mm (world coordinates).
                Default is [0.0, 0.0, 0.0]
           spacing (numpy array): 3D voxel spacing in mm. Default is [1,1,1]
           transform_matrix (numpy array): 3x3 image rotation matrix.
                Default is the 3x3 identity.
        """
        self._data = in_data
        if isinstance(self._data, bcolz.carray_ext.carray):
            self._mhd_param = OrderedDict(in_data.attrs["mhd_param"])
        else:
            self._mhd_param = MHD_DEFAULTS.copy()
        self._mhd_param["DimSize"] = array_to_string(
            np.array(in_data.shape[::-1], dtype=np.int64)
        )

        if offset is None:
            self._offset = np.zeros(3, dtype=np.float64)
        else:
            self._offset = offset

        if spacing is None:
            self._spacing = np.ones(3, dtype=np.float64)
        else:
            self._spacing = spacing

        if transform_matrix is None:
            self._transform_matrix = np.identity(3, dtype=np.float64)
        else:
            self._transform_matrix = transform_matrix

        self._voxel_to_world_matrix = np.identity(4, dtype=np.float64)
        self._world_to_voxel_matrix = np.identity(4, dtype=np.float64)

        self._update_geometry()

    def _update_geometry(self):
        """
        Initialise internal transformation matrices between world and
        voxel ccordinates from current offset, transformation matrix
        and spacing. If the current transformation matrix is not
        invertible, an exception is risen.
        Note that the mhd orientation is transposed according to
        the ITK conventions.
        """
        self._mhd_param["Offset"] = array_to_string(self._offset)
        self._mhd_param["TransformMatrix"] = array_to_string(
            self._transform_matrix.T
        )
        self._mhd_param["ElementSpacing"] = array_to_string(self._spacing)

        if (
            isinstance(self._data, bcolz.carray_ext.carray)
            and self._data.mode != "r"
        ):
            self._data.attrs["mhd_param"] = list(self._mhd_param.items())

        try:
            self._voxel_to_world_matrix.fill(0.0)
            self._voxel_to_world_matrix[0:3, 3] = self._offset
            self._voxel_to_world_matrix[3, 3] = 1
            self._voxel_to_world_matrix[:3, :3] = np.dot(
                self._transform_matrix, np.diag(self._spacing)
            )
            self._world_to_voxel_matrix = np.linalg.inv(
                self._voxel_to_world_matrix
            )
        except np.linalg.LinAlgError:
            raise ValueError("Illegal geometry settings")

    @property
    def mhd_param(self):
        """
        Getter for mhd parameters.

        Returns:
            OrderedDict mhd parameters
        """
        return self._mhd_param.copy()

    @property
    def offset(self):
        """
        Getter for 3D image offset.

        Returns:
            numpy array, length 3: Image offset in mm
        """
        return self._offset[:]

    @offset.setter
    def offset(self, new_offset):
        """
        Setter for 3D image offset.

        Changes offset and updates transformation matrices and
        bcolz.attribs for mhd export.

        Args:
            new_offset (numpy array): 3D offset in mm.
        """
        self._offset = new_offset[:]
        self._update_geometry()

    @property
    def transform_matrix(self):
        """
        Getter for image transformation matrix (rotation).

        Returns:
            3x3 numpy array: Image rotation matrix

        """
        return self._transform_matrix

    @transform_matrix.setter
    def transform_matrix(self, new_transform_matrix):
        """
            Setter for image rotation matrix.

            Changes rotation matrix and updates transformation matrices and
            bcolz.attribs for mhd export.

            Args:
                new_transform_matrix (3x3 numpy array): New rotation matrix.

        """
        self._transform_matrix = new_transform_matrix
        self._update_geometry()

    @property
    def spacing(self):
        """
        Getter for 3D image spacing.

        Returns:
            numpy array, length 3: Voxel size in mm
        """
        return self._spacing[:]

    @spacing.setter
    def spacing(self, new_spacing):
        """
        Setter for 3D image spacing.

        Changes spacing and updates transformation matrices and
        bcolz.attribs for mhd export.

        Args:
            new_spacing (numpy array): Voxel size in mm.
        """
        self._spacing = new_spacing
        self._update_geometry()

    @property
    def data(self):
        """
        Getter for 3D image data.

        Returns:
            bcolz.carray: Voxel size in mm
        """
        return self._data

    def clear_data(self):
        """
        Flush data to disk if required and free internal data
        Set data to None
        """
        if isinstance(self.data, bcolz.carray_ext.carray):
            if self.data.mode != "r":
                self.data.flush()
        # free data array
        del self._data

    @property
    def shape(self):
        """
        Getter for 3D image dimensions.

        Returns:
            tuple: Shape of 3D image data, order zyx (!)
        """
        return self._data.shape[:]

    @property
    def shape_array_xyz(self):
        """
        Getter for 3D image dimensions.

        Returns:
            numpy array: Shape of 3D image data, order xyz (!)
        """
        return np.array(self._data.shape[::-1], dtype=np.uint64)

    @property
    def nr_of_voxels(self):
        """
        Getter for 3D image number of voxels.

        Returns:
            python int: number of voxels
        """
        vox_num = 1
        for s in self._data.shape:
            vox_num *= int(s)
        return vox_num

    @property
    def mem_mb(self):
        """
        Getter for 3D image size in mb.

        Returns:
            int: Image size in mb
        """
        return 2 * self.nr_of_voxels // 2 ** 20

    @property
    def total_voxel_surface(self):
        """
        Calculate surface of all voxels

        Returns:
            float: total voxel surface in mm^2
        """

        single_voxel_surface = 2.0 * (self.spacing ** 2).sum()
        return self.nr_of_voxels * single_voxel_surface

    @property
    def world_to_voxel_matrix(self):
        """
        Getter for world to voxel coordinate transformation matrix.

        Returns:
            4x4 numpy array: World to voxel coordinate transformation
                matrix in homogeneous coordinates.
        """
        return self._world_to_voxel_matrix[:, :]

    @property
    def voxel_to_world_matrix(self):
        """
        Getter for voxel to world coordinate transformation matrix.

        Returns:
            4x4 numpy array: Voxel to world coordinate transformation
                matrix in homogeneous coordinates.
        """
        return self._voxel_to_world_matrix[:, :]

    def transform_voxel_to_world(self, vox_pos):
        """
        Transform voxel coordinate into world coordinate system.

        Args:
            vox_pos (numpy array, length 3): position in voxel coordinates
        Returns:
            numpy array, length 3: position in world coordinates
        """
        vox_pos_homogeneous = np.append(vox_pos, 1)
        return np.dot(self._voxel_to_world_matrix, vox_pos_homogeneous)[:3]

    def transform_world_to_voxel(self, world_pos):
        """
        Transform world coordinate into voxel coordinate system.

        Args:
            world_pos (numpy array, length 3): position in world coordinates
        Returns:
            numpy array, length 3: position in voxel coordinates
        """
        world_pos_homogeneous = np.append(world_pos, 1)
        return np.dot(self._world_to_voxel_matrix, world_pos_homogeneous)[:3]

    def uncompress(self):
        """
        Uncompress bcolz array in place. If the data is already uncompressed,
        the command does nothing.
        """
        if isinstance(self._data, bcolz.carray_ext.carray):
            self._data = self._data[:, :, :]


# data type for bcolz arrays
BCOLZ_DTYPE = np.int16

# Default values for mhd file parameters
MHD_DEFAULTS = OrderedDict(
    [
        ("ObjectType", "Image"),
        ("NDims", "3"),
        ("BinaryData", "True"),
        ("BinaryDataByteOrderMSB", "False"),
        ("CompressedData", "False"),
        ("TransformMatrix", "1 0 0 0 1 0 0 0 1"),
        ("Offset", "0 0 0"),
        ("CenterOfRotation", "0 0 0"),
        ("AnatomicalOrientation", "RAI"),
        ("ElementSpacing", "1 1 1"),
        ("DimSize", "1 1 1"),
        ("ElementType", "MET_SHORT"),
        ("HeaderSize", "0"),
        ("ElementDataFile", "binary.raw"),
    ]
)


def _calculate_header_size(raw_file_name, param):
    header_size = int(param["HeaderSize"])
    if header_size == -1:
        type_str = _MHD_TYPES[param["ElementType"]][1]
        dim_size = string_to_array(param["DimSize"], np.uint64)
        array_size_in_bytes = (
            int(np.prod(dim_size)) * np.dtype(type_str).itemsize
        )
        header_size = os.path.getsize(raw_file_name) - array_size_in_bytes
    return header_size


def read_mhd_param(in_file_name):
    """
    Read meta image header file into ordered dictionary

    Args:
        in_file_name (str): Name of meta image file including suffix.
    Returns:
        OrderedDict containing mhd settings from file.
    """
    mhd_param = MHD_DEFAULTS.copy()
    with open(in_file_name, "r") as in_file:
        for line in in_file:
            mhd_item = line[:-1].split("=")
            mhd_item[0] = mhd_item[0].strip()
            mhd_item[1] = mhd_item[1].strip()
            if mhd_item[0] in mhd_param.keys():
                mhd_param[mhd_item[0]] = mhd_item[1][:]
    return mhd_param


def write_mhd_file(mhd_param, out_file_name):
    with open(out_file_name, "w") as out_file:
        for i in mhd_param.items():
            out_file.write(i[0] + " = " + i[1] + "\n")


# conversion from mhd to numpy data types
_MHD_TYPES = {
    "MET_UCHAR": (np.uint8, "u1"),
    "MET_CHAR": (np.int8, "i1"),
    "MET_USHORT": (np.uint16, "u2"),
    "MET_SHORT": (np.int16, "i2"),
    "MET_UINT": (np.uint32, "u4"),
    "MET_INT": (np.int32, "i4"),
    "MET_ULONG": (np.uint64, "u8"),
    "MET_LONG": (np.int64, "i8"),
    "MET_FLOAT": (np.float32, "f4"),
    "MET_DOUBLE": (np.float64, "f8"),
}


def _raw_to_bcolz(
    raw_file_name,
    file_byteorder,
    dim_size,
    dtype,
    root_dir_name,
    n_bits,
    grey_range,
    is_mask,
    header_size,
):
    """
    Read contents of a binary file into a compressed bcolz array.

    A binary file containing grey values is read into a compressed bcolz
    array. Only grey values in the range specified by grey_range are used
    (clampling is applied). The number of bits to be used in the bcolz
    array can be specified resulting in an output grey value range from
    1 to 2**n_bits-1. If the byte order of the file is different from the
    system byte order, byte swapping is applied.

    Args:
        raw_file_name (str): Name of input file containing raw data
        file_byteorder (str): byte order of data in file ('big' or 'little')
        dim_size (numpy array): image dimensions
        dtype (numpy dtype): dtype of data in file
        root_dir_name (str): directory for bcolz on-disk array
        n_bits (int): number of bits to use for image grey values
        grey_range (tuple): minimum and maximum grey value to use
        is_mask (bool): flag for mask images
        header_size(int): header size in bytes
    """
    # input grey value range
    g_min, g_max = grey_range
    # output grey value range
    out_min, out_max = 1.0, 2.0 ** n_bits - 1.0
    # total number of voxels
    expected_len = int(dim_size[0]) * int(dim_size[1]) * int(dim_size[2])
    bcolz_array = None
    # voxels per slice
    swap_needed = file_byteorder != sys.byteorder
    # epsilon offset for rounding
    epsilon = np.finfo(np.float64).eps
    # read slices and append them to bcolz carray
    with open(raw_file_name, "r") as bin_file:
        # ignore header (size is given in byes)
        np.fromfile(bin_file, np.dtype("u1"), header_size)
        items_per_slice = dim_size[0] * dim_size[1]
        out_slice = np.zeros(
            shape=(1, dim_size[1], dim_size[0]), dtype=BCOLZ_DTYPE
        )
        for _ in range(dim_size[2]):
            im_slice = np.fromfile(bin_file, dtype, items_per_slice)
            im_slice = im_slice.reshape(1, dim_size[1], dim_size[0])
            if swap_needed:
                im_slice = im_slice.byteswap()
            im_slice = im_slice.astype(np.float64)
            rescale_gvals(
                im_slice,
                out_slice,
                g_min,
                g_max,
                out_min,
                out_max,
                epsilon,
                is_mask,
            )
            if bcolz_array is None:
                # create new bcolz array
                # use defaults defined in lcreg.py
                c_params = None
                if (2 * expected_len) // 2 ** 20 <= BCOLZ_COMPRESSION_LIMIT_MB:
                    c_params = bcolz.cparams(clevel=1, shuffle=bcolz.NOSHUFFLE)
                bcolz_array = bcolz.carray(
                    out_slice,
                    rootdir=root_dir_name,
                    mode="w",
                    expectedlen=expected_len,
                    cparams=c_params,
                )
            else:
                # append data
                bcolz_array.append(out_slice)

    # cleanup
    bcolz_array.flush()
    return bcolz_array


def open_image(root_dir_name, mode="r"):
    """
    Open bcolz-compressed image

    Args:
        root_dir_name (str): Name of bcolz root dir
        mode (str): Mode for bcolz file

    Returns:
        Image3D created from bcolz carray
    """
    bcolz_array = bcolz.open(root_dir_name, mode=mode)
    param = OrderedDict(bcolz_array.attrs["mhd_param"])
    offset = string_to_array(param["Offset"])
    spacing = string_to_array(param["ElementSpacing"])
    transform_matrix = (
        string_to_array(param["TransformMatrix"]).reshape((3, 3)).T
    )
    return Image3D(bcolz_array, offset, spacing, transform_matrix)


def close_image(image):
    """
    Flush image file, free internal cache and delete image

    Args:
        image (Image3D): Image to close
    """
    image.clear_data()
    del image
    gc.collect()


def import_image(
    in_file_name, root_dir_name, n_bits=None, grey_range=None, is_mask=False
):
    """
    Import mhd image into internal format based on compressed
    bcolz arrays.

    Args:
       in_file_name (str): Name of meta image file to import
       root_dir_name (str): Name of bcolz root dir
       n_bits (int): Number of bits to use for internal grey value
            representation.  Default is to determine bits from grey_range
       grey_range (2-tuple): range of grey values to be used in the bcolz
            array. Values outside this range are clipped
       is_mask (bool): Flag indicating whether image is a binary mask
    """
    # read image parameters
    if in_file_name[-3:].lower() == "mhd":
        param = read_mhd_param(in_file_name)
    else:
        raise TypeError("Unsupported image type")
    # compressed images are not supported
    if param["CompressedData"] == "True":
        raise TypeError("Compressed raw data is not supported")
    # no support for input file lists
    if (
        param["ElementDataFile"] == "LIST"
        or len(param["ElementDataFile"].split()) > 1
    ):
        raise TypeError("One-Slice-Per-File Data is not supported")
    dtype = _MHD_TYPES[param["ElementType"]][0]
    # get image dimensions
    dim_size = string_to_array(param["DimSize"], np.uint64)
    # check if byte swapping is required
    if param["BinaryDataByteOrderMSB"] == "True":
        data_byteorder = "big"
    else:
        data_byteorder = "little"
    swap_needed = data_byteorder != sys.byteorder
    data_bits = n_bits
    dir_name = os.path.dirname(in_file_name)
    raw_file_name = os.path.join(dir_name, param["ElementDataFile"])
    # calculate header size
    header_size = _calculate_header_size(raw_file_name, param)
    # for float input images, the default accuracy is 10 bit
    if dtype in (np.float32, np.float64):
        g_min = np.finfo(dtype).max
        g_max = np.finfo(dtype).min
        if data_bits is None:
            data_bits = 10
    else:
        g_min = np.iinfo(dtype).max
        g_max = np.iinfo(dtype).min
    if grey_range is not None:
        # apply user defined grey value range
        g_min, g_max = grey_range
    else:
        # determine grey value range from image data
        im_dtype = _MHD_TYPES[param["ElementType"]][1]
        if param["BinaryDataByteOrderMSB"] == "True":
            im_dtype = ">" + im_dtype
        else:
            im_dtype = "<" + im_dtype
        im_data = np.memmap(
            raw_file_name,
            mode="r",
            dtype=np.dtype(im_dtype),
            offset=header_size,
        )
        g_min = im_data.min()
        g_max = im_data.max()
        del im_data
    grey_range = (g_min, g_max)
    logging.debug(
        "{} gVal range: {} - {}, offset: {}".format(
            in_file_name, g_min, g_max, header_size
        )
    )

    if data_bits is None:
        # get Nr of bits from integer grey value range
        delta = 1 - np.finfo(np.float64).eps
        data_bits = int(np.log(g_max - g_min) / np.log(2) + delta)

    # convert raw data file into bcolz array
    bcolz_array = _raw_to_bcolz(
        raw_file_name,
        data_byteorder,
        dim_size,
        dtype,
        root_dir_name,
        data_bits,
        grey_range,
        is_mask,
        header_size,
    )
    # update mhd parameters
    param["ElementDataFile"] = os.path.basename(root_dir_name)
    param["CompressedData"] = "True"
    if sys.byteorder == "big":
        param["BinaryDataByteOrderMSB"] = "True"
    else:
        param["BinaryDataByteOrderMSB"] = "False"
    param["ElementType"] = "MET_SHORT"
    # store mhd parameters in bcolz attributes
    bcolz_array.attrs["mhd_param"] = list(param.items())
    # write first log entry for this image
    bcolz_array.attrs["log"] = [
        time.asctime() + ": " + 'imported from file "' + in_file_name + '"',
    ]
    if not is_mask:
        bcolz_array.attrs["lv_max"] = str((2 ** data_bits - 1) ** 2)
    # cleanup
    bcolz_array.flush()
    del bcolz_array


def export_image(im, file_name):
    """
    Export an image to mhd format.

    Args:
        im (Image3D): image
        file_name: name of output file
    """
    param = im.mhd_param.copy()
    param["ElementDataFile"] = os.path.basename(file_name)[:-4] + ".raw"
    write_mhd_file(param, file_name)
    data_file_name = os.path.join(
        os.path.dirname(file_name), param["ElementDataFile"]
    )
    with open(data_file_name, "wb") as out_file:
        out_file.write(im.data[:, :, :].copy())


def export_bcolz_image(root_dir_name, out_file_name, take_abs=False):
    """
    Export internal image representation to meta image

    Args:
        root_dir_name (str): bcolz root dir name
        out_file_name(str): name of output meta image including suffix
    """
    bcolz_array = bcolz.open(root_dir_name, mode="r")
    # get mhd parameters
    param = OrderedDict(bcolz_array.attrs["mhd_param"])
    param["CompressedData"] = "False"
    # name of mhd raw file
    param["ElementDataFile"] = os.path.basename(out_file_name)[:-3] + "raw"
    write_mhd_file(param, out_file_name)
    data_file_name = os.path.join(
        os.path.dirname(out_file_name), param["ElementDataFile"]
    )
    # store image data slice by slice into raw file
    with open(data_file_name, "wb") as out_file:
        out_slice = np.zeros(
            (bcolz_array.shape[1], bcolz_array.shape[2]), dtype=np.int16
        )
        for slice_ctr in range(bcolz_array.shape[0]):
            remove_rle(bcolz_array[slice_ctr, :, :], out_slice)
            if take_abs:
                out_file.write(np.abs(out_slice))
            else:
                out_file.write(out_slice)
    del bcolz_array


def export_bcolz_image_main():
    #
    # usage message
    if len(sys.argv) != 3:
        print("Export compressed image to mhd format")
        print("usage: compressed_to_mhd bcolz_dir_name out_mhd_name")
    else:
        export_bcolz_image(sys.argv[1], sys.argv[2])


def create_image(
    root_dir_name,
    dim_size,
    offset=None,
    spacing=None,
    transform_matrix=None,
    mode="w",
):
    """
    Create Image3D including a new compressed bcolz file.

    Args:
        root_dir_name (str): bcolz root dir name
        dim_size (numpy array): 3D image dimensions
        offset (numpy array): 3D offset (geometry)
        spacing (numpy array): 3D voxel size
        transform_matrix (3x3 numpy array): rotation matrix
        mode (str): image file mode

    Returns:
        Image3D image object
    """
    expected_len = int(dim_size[0]) * int(dim_size[1]) * int(dim_size[2])
    # create emtpy slice
    im_slice = np.zeros(shape=(1, dim_size[1], dim_size[2]), dtype=BCOLZ_DTYPE)

    # use default defined in lcreg.py
    c_params = None
    if (2 * expected_len) // 2 ** 20 <= BCOLZ_COMPRESSION_LIMIT_MB:
        c_params = bcolz.cparams(clevel=1, shuffle=bcolz.NOSHUFFLE)
    data = bcolz.carray(
        im_slice,
        rootdir=root_dir_name,
        cparams=c_params,
        expectedlen=expected_len,
        mode=mode,
    )
    # append missing sizes
    for _ in range(dim_size[0] - 1):
        data.append(im_slice)
    # create and fill in mhd parameters
    param = MHD_DEFAULTS.copy()
    param["ElementDataFile"] = os.path.basename(root_dir_name)
    param["DimSize"] = array_to_string(dim_size[::-1])
    if offset is not None:
        param["Offset"] = array_to_string(offset)
    if spacing is not None:
        param["ElementSpacing"] = array_to_string(spacing)
    if transform_matrix is not None:
        param["TransformMatrix"] = array_to_string(transform_matrix.T)
    data.attrs["mhd_param"] = list(param.items())
    # create log attribute
    data.attrs["log"] = [
        time.asctime()
        + ": "
        + 'image created with root_dir_name "'
        + root_dir_name
        + '", dim_size='
        + array_to_string(dim_size)
    ]
    data.flush()
    return Image3D(data, offset, spacing, transform_matrix)


def empty_image_like(template, root_dir_name):
    """
    Create an image with grey values set to zero from a template.

    Args:
        template (Image3D): template image defining image size and geometry.

    Returns:
        Image3D with same size and geometry as template but with all
        grey values set to zero.

    """
    return create_image(
        root_dir_name,
        template.data.shape,
        template.offset,
        template.spacing,
        template.transform_matrix,
    )


def image_part(source_im, offset_voxel, size_voxel):
    """
    Extract and uncompress a subimage.

    Args:
        source_im (Image3D): source to extract part from
        offset_voxel (numpy int array of length 3): 3D offset in voxels
        size_voxel   (numpy int array of length 3): 3D subimage size in voxels

    Returns:
        Image3D image with correct geometry information
    """
    # calculate new offset
    # (Position of voxel 0 of the image part in world coordinates
    new_world_offset = source_im.transform_voxel_to_world(offset_voxel)
    new_max_vox = offset_voxel + size_voxel
    new_data = source_im.data[
        offset_voxel[2] : new_max_vox[2],
        offset_voxel[1] : new_max_vox[1],
        offset_voxel[0] : new_max_vox[0],
    ]
    # create new image3d with adapted data and geometry
    return Image3D(
        new_data,
        new_world_offset,
        source_im.spacing,
        source_im.transform_matrix,
    )


def reduce_mhd(in_im_name, out_im_name, skip_3d):
    """
        Downsample an image by simply skipping grey values.

        Args:
            in_im_name (str): Name of input mhd image
            out_im_name (str): Name of output mhd image
            skip_3d (np.array): integer skipping factors, order: xyz
    """
    in_mhd = read_mhd_param(in_im_name)
    in_dtype = _MHD_TYPES[in_mhd["ElementType"]][0]
    in_shape = string_to_array(in_mhd["DimSize"], np.uint64)[::-1]
    in_data_file_name = os.path.join(
        os.path.dirname(in_im_name), in_mhd["ElementDataFile"]
    )
    in_data = np.memmap(
        in_data_file_name, dtype=in_dtype, mode="r", shape=tuple(in_shape)
    )
    out_shape = in_data[:: skip_3d[2], :: skip_3d[1], :: skip_3d[0]].shape
    out_spacing = string_to_array(in_mhd["ElementSpacing"]) * skip_3d
    out_mhd = in_mhd.copy()
    out_mhd["ElementSpacing"] = array_to_string(out_spacing)
    out_dtype = in_dtype
    out_data_name = os.path.basename(out_im_name)[:-3] + "raw"
    out_raw_name = os.path.join(os.path.dirname(out_im_name), out_data_name)
    out_mhd["ElementDataFile"] = os.path.basename(out_data_name)
    out_mhd["DimSize"] = array_to_string(out_shape[::-1])
    write_mhd_file(out_mhd, out_im_name)
    out_data = np.memmap(
        out_raw_name, dtype=out_dtype, shape=tuple(out_shape), mode="write"
    )
    out_data[:, :, :] = in_data[:: skip_3d[2], :: skip_3d[1], :: skip_3d[0]]
    out_data.flush()
    del in_data
    del out_data


def resample_mhd(
    fixed_mhd_file_name,
    moving_mhd_file_name,
    out_mhd_file_name,
    matrix,
    offset,
):
    """
        Resample an mhd image using scipy and np.memmap

        Images do not need to fit into memory

        Args:
            fixed_mhd_file_name (str): name of fixed image mhd file
            moving_mhd_file_name (str): name of moving image mhd file
            out_mhd_file_name (str): name of output image mhd file
            matrix (3x3 np.array): rotation matrix
            offset (1x3 np.array): offset vector
    """
    try:
        from scipy.ndimage import affine_transform
    except ImportError:
        logging.error("scipy not installaed")
        print("image resampling requires scipy")
        return
    # read mhd files
    fixed_mhd = read_mhd_param(fixed_mhd_file_name)
    out_mhd = fixed_mhd.copy()
    moving_mhd = read_mhd_param(moving_mhd_file_name)
    # get shapes and dtypes
    out_shape = string_to_array(out_mhd["DimSize"], np.uint64)[::-1]
    out_dtype = _MHD_TYPES[out_mhd["ElementType"]][1]
    if out_mhd["BinaryDataByteOrderMSB"] == "True":
        out_dtype = ">" + out_dtype
    else:
        out_dtype = "<" + out_dtype
    moving_shape = string_to_array(moving_mhd["DimSize"], np.uint64)[::-1]
    moving_dtype = _MHD_TYPES[moving_mhd["ElementType"]][1]
    if moving_mhd["BinaryDataByteOrderMSB"] == "True":
        moving_dtype = ">" + moving_dtype
    else:
        moving_dtype = "<" + moving_dtype
    # update and store output mhd parameters
    out_data_name = os.path.basename(out_mhd_file_name)[:-3] + "raw"
    out_raw_name = os.path.join(
        os.path.dirname(out_mhd_file_name), out_data_name
    )
    out_mhd["ElementDataFile"] = out_data_name
    out_mhd["HeaderSize"] = "0"
    write_mhd_file(out_mhd, out_mhd_file_name)
    # open mmaps
    moving_data_file_name = os.path.join(
        os.path.dirname(moving_mhd_file_name), moving_mhd["ElementDataFile"]
    )
    moving_offset = _calculate_header_size(moving_data_file_name, moving_mhd)
    moving_data = np.memmap(
        moving_data_file_name,
        dtype=np.dtype(moving_dtype),
        mode="r",
        shape=tuple(moving_shape),
        offset=moving_offset,
    )
    out_data = np.memmap(
        out_raw_name,
        dtype=np.dtype(out_dtype),
        shape=tuple(out_shape),
        mode="write",
        offset=0,
    )
    # perform transform using linear interpolation
    affine_transform(
        moving_data,
        matrix,
        offset,
        output_shape=out_data.shape,
        output=out_data,
        order=1,
        mode="constant",
        cval=0,
        prefilter=False,
    )
    # cleanup
    out_data.flush()
    del moving_data
    del out_data


def difference_image_mhd(
    im1_mhd_file_name, im2_mhd_file_name, out_mhd_file_name, calculate_abs=True
):
    """
        Calculate absolute voxel by voxel difference betwen two images.

        Image sizes must agree but images need not to fit into memory.

        Args:
            im1_mhd_file_name (str): name of first image mhd file
            im2_mhd_file_name (str): name of second image mhd file
            out_mhd_file_name (str): name of output image mhd file
            calculate_abs (boolean): flag for absolute value calculation
    """
    # read mhd files
    im1_mhd = read_mhd_param(im1_mhd_file_name)
    out_mhd = im1_mhd.copy()
    im2_mhd = read_mhd_param(im2_mhd_file_name)
    # get shapes and dtypes
    out_shape = string_to_array(out_mhd["DimSize"], np.uint64)[::-1]
    out_dtype = _MHD_TYPES[out_mhd["ElementType"]][1]
    if out_mhd["BinaryDataByteOrderMSB"] == "True":
        out_dtype = ">" + out_dtype
    else:
        out_dtype = "<" + out_dtype
    im1_shape = string_to_array(im1_mhd["DimSize"], np.uint64)[::-1]
    im1_dtype = _MHD_TYPES[im1_mhd["ElementType"]][1]
    if im1_mhd["BinaryDataByteOrderMSB"] == "True":
        im1_dtype = ">" + im1_dtype
    else:
        im1_dtype = "<" + im1_dtype
    im2_shape = string_to_array(im2_mhd["DimSize"], np.uint64)[::-1]
    im2_dtype = _MHD_TYPES[im2_mhd["ElementType"]][1]
    if im2_mhd["BinaryDataByteOrderMSB"] == "True":
        im2_dtype = ">" + im2_dtype
    else:
        im2_dtype = "<" + im2_dtype
    # check sizes
    if not np.array_equal(im2_shape, im1_shape):
        print("ERROR: Image sizes must agree\n")
        sys.exit(0)
    else:
        # update and store output mhd parameters
        out_data_name = os.path.basename(out_mhd_file_name)[:-3] + "raw"
        out_raw_name = os.path.join(
            os.path.dirname(out_mhd_file_name), out_data_name
        )
        out_mhd["ElementDataFile"] = out_data_name
        out_mhd["HeaderSize"] = "0"
        write_mhd_file(out_mhd, out_mhd_file_name)
        # open mmaps
        im1_data_file_name = os.path.join(
            os.path.dirname(im1_mhd_file_name), im1_mhd["ElementDataFile"]
        )
        im1_offset = _calculate_header_size(im1_data_file_name, im1_mhd)
        im1_data = np.memmap(
            im1_data_file_name,
            dtype=np.dtype(im1_dtype),
            mode="r",
            shape=tuple(im1_shape),
            offset=im1_offset,
        )
        im2_data_file_name = os.path.join(
            os.path.dirname(im2_mhd_file_name), im2_mhd["ElementDataFile"]
        )
        im2_offset = _calculate_header_size(im2_data_file_name, im2_mhd)
        im2_data = np.memmap(
            im2_data_file_name,
            dtype=np.dtype(im2_dtype),
            mode="r",
            shape=tuple(im2_shape),
            offset=im2_offset,
        )
        out_data = np.memmap(
            out_raw_name,
            dtype=np.dtype(out_dtype),
            shape=tuple(out_shape),
            mode="write",
            offset=0,
        )
        # calculate difference
        if calculate_abs:
            np.abs(im1_data - im2_data, out=out_data)
        else:
            np.subtract(im1_data, im2_data, out=out_data)
        # cleanup
        out_data.flush()
        del im1_data
        del im2_data
        del out_data


def abs_difference_mhd_main():
    #
    # usage message
    if len(sys.argv) < 3:
        print("calculate absolute difference between two mhd images")
        print("usage: abs_difference_mhd im1_name im2_name out_im_name")
    else:
        difference_image_mhd(
            sys.argv[1], sys.argv[2], sys.argv[3], calculate_abs=True
        )


def difference_mhd_main():
    #
    # usage message
    if len(sys.argv) < 3:
        print("calculate signed difference between two mhd images")
        print("usage: difference_mhd im1_name im2_name out_im_name")
    else:
        difference_image_mhd(
            sys.argv[1], sys.argv[2], sys.argv[3], calculate_abs=False
        )



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

name = "lcreg"
__version__ = "1.0.0"
